//use lazy_static::lazy_static;
//use tfhe::shortint::parameters::{PARAM_MESSAGE_1_CARRY_0_KS_PBS, ShortintParameterSet};
use tfhe::shortint::parameters::{ShortintParameterSet};




//lazy_static! {
//    pub static ref PARAMS: ShortintParameterSet = PARAM_MESSAGE_1_CARRY_0_KS_PBS.try_into().unwrap();
//}

//use tfhe::shortint::parameters::{LEGACY_WOPBS_PARAM_MESSAGE_4_CARRY_0_KS_PBS,LEGACY_WOPBS_PARAM_MESSAGE_4_CARRY_1_KS_PBS,LEGACY_WOPBS_PARAM_MESSAGE_1_CARRY_0_KS_PBS,LEGACY_WOPBS_PARAM_MESSAGE_2_CARRY_0_KS_PBS};
